import autosubmitAPIwu.experiment.common_requests as ExperimentUtils


def main():
    """
    Process and positions graph drawing.
    """
    ExperimentUtils.process_active_graphs()


if __name__ == "__main__":
    main()
