import autosubmitAPIwu.experiment.common_requests as ExperimentUtils


def main():
    # Process queue and running times
    ExperimentUtils.process_completed_times()


if __name__ == "__main__":
    main()
